using Newtonsoft.Json;
using System;
using System.Collections;
using System.Numerics;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using GrapeCity.Documents.Excel;
using GrapeCity.Documents.Excel.Expressions;
using GrapeCity.Documents.Excel.Drawing;
using System.Globalization;

namespace GrapeCity.Documents.Excel.Examples.Features.Formulas.Expressions.ModifyFormula
{
    class Program
    {
        static void Main(string[] args)
        {
            //create a new workbook
            var workbook = new GrapeCity.Documents.Excel.Workbook();

            //Enable dynamic array formula
            workbook.AllowDynamicArray = true;

            //Load sample data from excel file
            workbook.Open("SampleData.xlsx");

            //Fetch worksheet
            var worksheet = workbook.Worksheets[0];

            //Apply styling to sample data
            worksheet.Range["A1:D1, H1:K1"].Interior.Color = System.Drawing.Color.FromArgb(68, 114, 196);
            worksheet.Range["A1:D1, H1:K1"].Font.Color = System.Drawing.Color.White;
            worksheet.Range["A2:D19, H3#"].Borders.Color = System.Drawing.Color.FromArgb(91, 155, 213);
            worksheet.Range["A2:D19, H3#"].Borders.LineStyle = BorderLineStyle.Thin;

            //Fetch the original formula which needs to be parsed.            
            var originalFormula = worksheet.Range["H3"].Formula.Substring(1);

            //Add new worksheet to print syntax tree
            var treeWorksheet = workbook.Worksheets.Add();

            //Invoke method to parse and print formula syntax tree
            ParseModify.ParseAndPrint(treeWorksheet, originalFormula);

            //Invoke method to parse and modify formula
            ParseModify.ModifyFormula(worksheet, originalFormula);

            //Save modified Excel file
            workbook.Save("ModifiedFormula.xlsx", SaveFileFormat.Xlsx);
        }
    }

    class ParseModify
    {
        //Method to parse a formula and print the syntax tree
        public static void ParseAndPrint(IWorksheet worksheet, string formula)
        {
            // Get syntax tree
            var syntaxTree = FormulaSyntaxTree.Parse(formula);

            // Flatten nodes
            var displayItems = new List<(string TypeName, int IndentLevel, string Content)>();

            void flatten(SyntaxNode node, int level)
            {
                displayItems.Add((node.GetType().Name, level, node.ToString()));
                foreach (var child in node.Children)
                {
                    flatten(child, level + 1);
                }
            }

            flatten(syntaxTree.Root, 0);

            // Output           
            worksheet.ShowRowOutline = false;
            worksheet.OutlineColumn.ColumnIndex = 1;

            // Header
            worksheet.Range["A1"].Value = "Formula";
            worksheet.Range["A3"].Value = "Syntax node";
            worksheet.Range["B3"].Value = "Part";

            // Values
            worksheet.Range["B1"].Value = "'=" + formula;
            for (var i = 0; i < displayItems.Count; i++)
            {
                var item = displayItems[i];
                var text = "'" + item.TypeName;

                worksheet.Range[i + 4, 0].Value = text;
                worksheet.Range[i + 4, 0].IndentLevel = item.IndentLevel;
                worksheet.Range[i + 4, 1].Value = "'" + item.Content;
            }

            //Apply styling
            worksheet.Range["A1:B3"].Interior.Color = System.Drawing.Color.FromArgb(68, 114, 196);
            worksheet.Range["A1:B3"].Font.Color = System.Drawing.Color.White;
            worksheet.Range["A1:B3"].Borders.Color = System.Drawing.Color.FromArgb(91, 155, 213);
            worksheet.Range["A1:B3"].Borders.LineStyle = BorderLineStyle.Thin;
            worksheet.Range["A1,A3,B3"].Font.Size = 14;
            worksheet.Range["A1,A3,B3"].Font.Bold = true;
            worksheet.Range["A:C"].EntireColumn.AutoFit();
        }

        //Method to parse and modify the formula
        public static void ModifyFormula(IWorksheet worksheet, string originalFormula)
        {
            //Apply UNIQUE formula to get unique sales representatives list
            worksheet.Range["F1"].Value = "Unique Rep";
            worksheet.Range["F2"].Formula = "=UNIQUE(A2:A19)";
            var uniqueRep = worksheet.Range["F2#"];
            // Apply Styling
            worksheet.Range["F:F"].EntireColumn.AutoFit();
            worksheet.Range["F1"].Interior.Color = System.Drawing.Color.FromArgb(68, 114, 196);
            worksheet.Range["F1"].Font.Color = System.Drawing.Color.White;
            worksheet.Range["F2#"].Borders.Color = System.Drawing.Color.FromArgb(91, 155, 213);
            worksheet.Range["F2#"].Borders.LineStyle = BorderLineStyle.Thin;

            //Get syntax tree
            var syntaxTree = FormulaSyntaxTree.Parse(originalFormula);

            //Find
            var findText = new TextNode("Fritz");

            //Replacement
            var replaceText = new TextNode("");

            //Loop through names list to modify the formula for each sales representative
            for (int r = 0, resultRow = 3; r < uniqueRep.Cells.Count; r++, resultRow = resultRow + 4)
            {
                //Get name to be replaced in the formula
                var cval = uniqueRep.Cells[r].Value.ToString();

                if (findText.Value != cval)
                {
                    //Assign name to be replaced to Replace TextNode
                    replaceText.Value = cval;

                    //Invoke the recursive method to perform find and replace operation
                    replaceNode(syntaxTree.Root, findText, replaceText);

                    //Assign the modified formula to a cell in the worksheet
                    var resultRange = "H" + resultRow.ToString();
                    worksheet.Range[resultRange].Formula = "=" + syntaxTree.ToString();
                    worksheet.Range[resultRange + "#"].Borders.Color = System.Drawing.Color.FromArgb(91, 155, 213);
                    worksheet.Range[resultRange + "#"].Borders.LineStyle = BorderLineStyle.Thin;

                    //Update the value of Find node to perform find and replace operation for next sales representative name
                    findText = replaceText;
                }
            }

            //Find and replace
            void replaceNode(SyntaxNode lookIn, SyntaxNode find, SyntaxNode replacement)
            {
                var children = lookIn.Children;

                for (var i = 0; i < children.Count; i++)
                {
                    var child = children[i];
                    if (child.Equals(find))
                    {
                        children[i] = replacement;
                    }
                    else
                    {
                        replaceNode(child, find, replacement);
                    }
                }
            }
        }
    }
}